﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Xml.Linq;

namespace NillToMinOccurs
{
	// *************************************************************************
	// *** Program *************************************************************
	// *************************************************************************
	public class Program
	{
		// ---------------------------------------------------------------------
		// --- Program Methods -------------------------------------------------
		// ---------------------------------------------------------------------
		public static void Main(string[] args)
		{
			try // catch
			{
				Console.WriteLine("Elmhurst NILL XSD converter (c) 2011-13 v0.4");
				Console.WriteLine();

				// Help (MAN)
				if (args == null || args.Count() == 0)
				{
					Console.WriteLine("Examples:");
					Console.WriteLine();
					Console.WriteLine("Replace nil to min occurs (Fixing schema generated with XSD.EXE from C#)");
					Console.WriteLine("NILLTOMINOCCURS [File Name.XSD]");
					Console.WriteLine();
					Console.WriteLine("Add nil from min occurs 0 (Fixing general schema converted to C# by XSD.EXE)");
					Console.WriteLine("NILLTOMINOCCURS [File Name.XSD] -r");
					Console.WriteLine();
					Console.WriteLine("Linking schema1..?.xsd to schema0.xsd for different namespaces");
					Console.WriteLine("NILLTOMINOCCURS -sl");
					return;
				} // if

				// Identify operation
				if (args.Length == 1)
				{
					if (args[0] == "-sl")
						LinkDifferentNameSpaceSchemes();
					else
						CancelNillable(args[0]);
				} // if
				else if (args.Length == 2 && args[1] == "-r")
					RestoreNillable(args[0]);
				else
					throw new NotSupportedException("Invalid parameter(s)!");
			} // try
			catch (Exception e)
			{
				Console.WriteLine("Error: " + e.Message);
			} // catch
		}

		// ---------------------------------------------------------------------
		private static void CancelNillable(string fileName)
		{
			CancelNillable(fileName, new List<string>());
		}

		// ---------------------------------------------------------------------
		private static void CancelNillable(string fileName, List<string> namespacesDone)
		{
			if (namespacesDone == null)
				throw new NotSupportedException();

			// Schema file has been updated
			if (namespacesDone.Count(item => item.EndsWith(Path.GetFileName(fileName))) > 0)
				return;

			// Fix current XSD
			var xsdFile = new XsdFile(fileName);
			foreach (var complexElement in xsdFile.GetComplexTypeElements)
			{
				foreach (XElement element in complexElement.Elements())
				{
					var minOccursAttribute = element.Attributes().FirstOrDefault(item => "minOccurs".Equals(item.Name.ToString(), StringComparison.InvariantCulture));
					var nillableAttribute = element.Attributes().FirstOrDefault(item => "nillable".Equals(item.Name.ToString(), StringComparison.InvariantCulture));

					if (nillableAttribute != null && "true".Equals(nillableAttribute.Value, StringComparison.InvariantCulture))
					{
						if (minOccursAttribute != null)
							minOccursAttribute.Value = "0";
						else
							element.Add(new XAttribute("minOccurs", "0"));

						nillableAttribute.Remove();
					} // if
				} // foreach
			} // foreach

			// Save final XML
			xsdFile.Save();

			// Check linked XSD files
			namespacesDone.Add(fileName);
			foreach (var importElement in xsdFile.ImportedNamespaceElements)
			{
				var linkedFileName = XsdFile.GetFileLocation(importElement);
				CancelNillable(linkedFileName, namespacesDone);
			} // foreach
		}

		// ---------------------------------------------------------------------
		private static void RestoreNillable(string fileName)
		{
			RestoreNillable(fileName, new List<string>());
		}

		// ---------------------------------------------------------------------
		private static void RestoreNillable(string fileName, List<string> namespacesDone)
		{
			if (namespacesDone == null)
				throw new NotSupportedException();

			// Schema file has been updated
			if (namespacesDone.Count(item => item.EndsWith(Path.GetFileName(fileName))) > 0)
				return;

			var xsdFile = new XsdFile(fileName);
			foreach (var complexElement in xsdFile.GetComplexTypeElements)
			{
				foreach (XElement element in complexElement.Elements())
				{
					var minOccursAttribute = element.Attributes().FirstOrDefault(item => "minOccurs".Equals(item.Name.ToString(), StringComparison.InvariantCulture));
					var maxOccursAttribute = element.Attributes().FirstOrDefault(item => "maxOccurs".Equals(item.Name.ToString(), StringComparison.InvariantCulture));

					var nillableAttribute = element.Attributes().FirstOrDefault(item => "nillable".Equals(item.Name.ToString(), StringComparison.InvariantCulture));
					var typeAttribute = element.Attributes().FirstOrDefault(item => "type".Equals(item.Name.ToString(), StringComparison.InvariantCulture));

					if (minOccursAttribute != null && minOccursAttribute.Value == "0")
					{
						if (nillableAttribute != null)
							nillableAttribute.Value = "true";
						else
							element.Add(new XAttribute("nillable", "true"));
					} // if

					// Rmoeve min / max occurs, otherwise XSD generates *Specified Properties which break serialization to XML
					bool isSimplifiedType =
						typeAttribute != null &&
						(typeAttribute.Value ?? string.Empty).ToLower().Contains("xs:");	// Ordinal types (string, double, ...)

					// Enums (non pointers)
					if (typeAttribute != null && !isSimplifiedType)
					{
						var typeDefinitionElement = xsdFile.Root.Elements().FirstOrDefault(item => IsElementAttributeName(item, typeAttribute.Value));
						isSimplifiedType = (typeDefinitionElement != null && "simpleType".Equals(typeDefinitionElement.Name.LocalName, StringComparison.InvariantCulture));
					} // if

					bool isArray = (maxOccursAttribute == null || "unbounded".Equals(maxOccursAttribute.Value, StringComparison.InvariantCulture));

					if (isSimplifiedType && !isArray)
					{
						if (minOccursAttribute != null)
							minOccursAttribute.Remove();

						if (maxOccursAttribute != null)
							maxOccursAttribute.Remove();
					} // if
				} // foreach
			} // if

			// Save final XML
			xsdFile.Save();

			// Check linked XSD files
			namespacesDone.Add(fileName);
			foreach (var importElement in xsdFile.ImportedNamespaceElements)
			{
				var linkedFileName = XsdFile.GetFileLocation(importElement);
				RestoreNillable(linkedFileName, namespacesDone);
			} // foreach
		}

		// ---------------------------------------------------------------------
		private static bool IsElementAttributeName(XElement element, string name)
		{
			if (element == null)
				throw new ArgumentNullException("element");
			if (string.IsNullOrEmpty(name))
				throw new ArgumentNullException("name");

			var nameAttribute = element.Attributes().FirstOrDefault(attr => "name".Equals(attr.Name.ToString(), StringComparison.InvariantCulture));	// Has name element?
			return (nameAttribute != null && nameAttribute.Value.Equals(name, StringComparison.InvariantCulture));
		}

		// ---------------------------------------------------------------------
		private static void LinkDifferentNameSpaceSchemes()
		{
			const string MainSchemaFileName = "schema0.xsd";

			var schemes = Directory.GetFiles(/*path*/Environment.CurrentDirectory, /*pattern*/"schema*.xsd");
			
			if (schemes.Count() == 0)
				throw new Exception("Schemes files 'schema*.xsd' not found!");
			if (schemes.Count() == 1)
				throw new Exception("Cannot merge one file!");
			if (schemes.Count(item => item.EndsWith(MainSchemaFileName)) == 0)
				throw new Exception("Cannot find " + MainSchemaFileName);

			// Load main schema file and add link to external XSD files
			var mainSchema = new XsdFile(schemes.First(item => item.EndsWith(MainSchemaFileName)));
			foreach (var importElement in mainSchema.ImportedNamespaceElements)
			{
				string suggestedName = XsdFile.SuggestedFileName(importElement);
				XsdFile.SetFileLocation(importElement, suggestedName);
			} // foreach

			// Rename external XSDs
			foreach (string schemeFileName in schemes)
			{
				if (!schemeFileName.EndsWith(MainSchemaFileName))
				{
					var currentXsd = new XsdFile(schemeFileName);
					string newFileName = Path.Combine(Path.GetDirectoryName(schemeFileName), currentXsd.SuggestedFileName());

					if (File.Exists(newFileName))
						File.Delete(newFileName);

					File.Move(schemeFileName, newFileName);
				} // if
			} // foreach

			mainSchema.Save();
		}
	} // public class Program

} // namespace NillToMinOccurs