﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text.RegularExpressions;
using System.Xml.Linq;

namespace NillToMinOccurs
{
	// *************************************************************************
	// *** XsdFile *************************************************************
	// *************************************************************************
	public class XsdFile
	{
		// ---------------------------------------------------------------------
		private const string NameSpaceXS = "http://www.w3.org/2001/XMLSchema";
		private const string ImportSchemaLocationAttributeName = "schemaLocation";
		// ---------------------------------------------------------------------
		private readonly string fileName;
		private readonly XDocument xml;

		// ---------------------------------------------------------------------
		// --- XsdFile Constructor ---------------------------------------------
		// ---------------------------------------------------------------------
		public XsdFile(string fileName)
		{
			xml = XDocument.Load(fileName);
			this.fileName = fileName;
		}

		// ---------------------------------------------------------------------
		// --- XsdFile Properties ----------------------------------------------
		// ---------------------------------------------------------------------
		public string TargetNameSpace
		{
			get
			{
				var targetAttribute = xml.Root.Attributes().FirstOrDefault(item => "targetNamespace".Equals(item.Name.LocalName, StringComparison.InvariantCultureIgnoreCase));
				return targetAttribute != null ? targetAttribute.Value : null;
			} // get
		}

		// ---------------------------------------------------------------------
		public IEnumerable<XElement> ImportedNamespaceElements
		{
			get { return xml.Root.Elements().Where(item => "import".Equals(item.Name.LocalName, StringComparison.InvariantCultureIgnoreCase)); } // get
		}

		// ---------------------------------------------------------------------
		public IEnumerable<XElement> GetComplexTypeElements
		{
			get { return xml.Descendants(XName.Get("sequence", NameSpaceXS)); }
		}

		// ---------------------------------------------------------------------
		public XElement Root
		{
			get { return xml.Root; }
		}

		// ---------------------------------------------------------------------
		// --- XsdFile Methods -------------------------------------------------
		// ---------------------------------------------------------------------
		public void Save()
		{
			xml.Save(fileName);
		}

		// ---------------------------------------------------------------------
		public string SuggestedFileName()
		{
			return SuggestedFileName(TargetNameSpace);
		}

		// ---------------------------------------------------------------------
		public static string SuggestedFileName(XElement importElement)
		{
			if (importElement == null)
				throw new NotSupportedException();

			var namespaceAttribute = importElement.Attributes().FirstOrDefault(item => "namespace".Equals(item.Name.LocalName, StringComparison.InvariantCultureIgnoreCase));
			if (namespaceAttribute == null)
				throw new NotSupportedException();

			return SuggestedFileName(namespaceAttribute.Value);
		}

		// ---------------------------------------------------------------------
		public static string SuggestedFileName(string namespaceValue)
		{
			if (string.IsNullOrEmpty(namespaceValue))
				throw new NotSupportedException();

			var updatedNameSpace = namespaceValue.Replace("https://", string.Empty).Replace("http://", string.Empty);

			string regexSearch = new string(Path.GetInvalidFileNameChars()) + new string(Path.GetInvalidPathChars());
			var regex = new Regex(string.Format("[{0}]", Regex.Escape(regexSearch)));
			
			return regex.Replace(updatedNameSpace, string.Empty) + ".xsd";
		}

		// ---------------------------------------------------------------------
		public static void SetFileLocation(XElement importElement, string fileLocation)
		{
			if (importElement == null)
				return;

			var attribute = importElement.Attributes().FirstOrDefault(item => ImportSchemaLocationAttributeName.ToLower() == item.Name.LocalName.ToLower());
			if (attribute == null)
				importElement.Add(new XAttribute(ImportSchemaLocationAttributeName, fileLocation));
			else
				attribute.Value = fileLocation;
		}

		// ---------------------------------------------------------------------
		public static string GetFileLocation(XElement importElement)
		{
			if (importElement == null)
				return null;

			var attribute = importElement.Attributes().FirstOrDefault(item => ImportSchemaLocationAttributeName.ToLower() == item.Name.LocalName.ToLower());
			if (attribute == null)
				return null;

			return attribute.Value;
		}
	} // public class XsdFile

} // namespace NillToMinOccurs